// ***********************************************************************
// VISUAL MARRIAGE Applet
//
// Visualizes the algorithm for finding a man-optimal stable marriage.
// The simulated algorithm is the one of Gale and Shapley.
//
// by Rolf Bruderer, Oct 2003
// ***********************************************************************

import java.awt.*;

public class Man
	// Men that do some movements in a world.
extends Person 
{

	// man is a final state machine with following states.
	private static final int STATE_MOVE = 0;		// on the way to a proposal: walk up to border of woman.
	private static final int STATE_PROPOSAL = 1;	// doing proposal: walking nearer to the woman.
	private static final int STATE_MARRIED = 2;	    // married: just stay with your partner.
	private static final int STATE_WALKAWAY = 3;  	// walking away after denied proposal or divorcing.
	
	public Man (int xp, int yp)
	// Create and initialize
	// with x and y position `xp' and `yp'.
	{
		x = xp;
		y = yp;
		actionx = xp;
		actiony = yp;
		state = STATE_MOVE;	
		target_index=0;
		doingProposal = false;		
		actionInLastStep = false;
		married = false;
		partner = null;
	}
	
	public void reset()
	// restart man searching a woman.
	{
		super.reset();
		actionx = x;
		actiony = y;
		state = STATE_MOVE;	
		target_index=0;
		doingProposal = false;
		actionInLastStep = false;
	}
	
	private double actionx, actiony;
	// Coordinates where this man last started from.			

	private int state;  
	// the man's state of action/movement. (see step)
	
	private int target_index;
	// preference index of the woman this man actualy is focused on.

	private boolean actionInLastStep;
	// was there a relevant action step?
		
	private boolean doingProposal;
	// is this man the man who is right now doing a proposal
	// (if so he is responsible to endProposal when proposal phase has finished)
	
	public boolean occuredAction() 
	// did the man a action (=proposal) recently?
	{
		return actionInLastStep;
	}
	
	public void paint (Graphics g)
	// Paint this person in a world
	// with Graphics `g' at its position.
	{
		// paint the person as filled oval in his color.
		g.setColor(color);
		g.fillOval((int)x, (int)y, (int)size, (int)size);					
	}

	
	public void divorce ()
	// divorce this man from his woman.
	{
		state = STATE_WALKAWAY;
		doingProposal = false;
		preferences.setHighlightColor(partner, Color.red);		
		ownerTable.fireTableRowsUpdated(ownerTableRow, ownerTableRow);
		history_window.addDivorceMessage(this, partner);
		partner = null;
	}	
	
	private void move(double len) 
	// make move of length len into direction of woman.
	// (or move away if len is negative)
	{

		// Get the woman the man is actualy focused on.
		Woman target = (Woman) preferences.elementAt (target_index);

		// Calculate distance and distance vector in world to woman.		
		double vect_x, vect_y, distance;
		vect_x = target.x - x;
		vect_y = target.y - y;
		distance = Math.sqrt (vect_x * vect_x + vect_y * vect_y);
		
		// if very near directly go to target.
		if ((len > 0.0) && (distance < len + 1.0)) {
			x = target.x;
			y = target.y;				
			return;
		}
		
		// choose vector to move on (either man->target or actionpoint->target).
		double linelen = distance;
		vect_x = - vect_x;
		vect_y = - vect_y;
		if (distance < 2.0) {
			vect_x = actionx - target.x;
			vect_y = actiony - target.y;
			linelen = Math.sqrt (vect_x * vect_x + vect_y * vect_y);			
		}		
		
		// calculate new position in vector direction.
		distance = distance - len;
		x = target.x + ((vect_x / linelen) * distance);
		y = target.y + ((vect_y / linelen) * distance);
				
	}
		
	public void step()
	// let the man do the next step of its movement.
	// (implementation of final state machine)
	{
		
		// set 'actionInLastStep' to false by default.
		actionInLastStep = false;
		
		// Get the woman the man is actualy focused on.
		Woman target = (Woman) preferences.elementAt (target_index);
		
		// Calculate distance and distance vector in world to woman.		
		double vect_x, vect_y, distance;
		vect_x = target.x - x;
		vect_y = target.y - y;
		distance = Math.sqrt (vect_x * vect_x + vect_y * vect_y);

		// perform action according to state.
		switch (state) {		
			case STATE_MOVE:

				// only allowed to move, if no pending action.
				if (pendingAction) {
					return;
				}
				
				// near enough to do a proposal?
				if (distance <= 2*size) {
					state = STATE_PROPOSAL;
					target.startProposal(this);
					actionInLastStep = true;
					doingProposal = true;
					preferences.setBackgroundColor(target, Color.orange);
					ownerTable.fireTableRowsUpdated(ownerTableRow, ownerTableRow);
				}
				
				// otherwise move fast.
				else {
					move(size * speed / 3.0);				
				}
				
				break;
				
			case STATE_PROPOSAL:
			
				// JUST FOR DEBUGGING:
				// System.err.println(name + " : STATE_WALKAWAY (" + x + ", " + y + ")");				
								
				// if too far, move nearer before asking her.
				if (distance > size) {
					move(size * slow_speed / 3.0);
				}
				
				// otherwise ask her.
				else {
		
					if (target.acceptProposal (this)) {
						// marry her.
						state = STATE_MARRIED;
						married = true;
						partner = target;
						preferences.setHighlightColor(target, Color.yellow);
						preferences.setBackgroundColor(target, null);
						ownerTable.fireTableRowsUpdated(ownerTableRow, ownerTableRow);
					}
					else {
						// walk away.
						preferences.setHighlightColor(target, Color.red);
						preferences.setBackgroundColor(target, null);
						ownerTable.fireTableRowsUpdated(ownerTableRow, ownerTableRow);
						state = STATE_WALKAWAY;
					}
					
				}				
			
				break;
				
			case STATE_MARRIED:						
							
				// move to wife and stay there.
				if (distance > 0.0) {
					// move one step into direction of woman.
					move(size * slow_speed / 3.0);  
				}
				else {
					// check to finish proposal phase if responsible for it (doingProposal).
					if (doingProposal) {
						target.endProposal();
						doingProposal = false;
					}
				}
				break;
				
			case STATE_WALKAWAY:
			
				// JUST FOR DEBUGGING:
				// System.err.println(name + " : STATE_WALKAWAY (" + x + ", " + y + ")");				
			
				// just walk away for a little while (until you forget her).
				move(- size * slow_speed / 3.0);   // just make one step away from woman.
				
				// if far enough, search for next woman.
				if (distance > 2.0*size) {
					state = STATE_MOVE;
					target.endProposal();					
					actionx = target.x;
					actiony = target.y;
					target_index++;														
				}			
				break;
				
			default:
				// invalid state, just go to moving state.
				state = STATE_MOVE;
		
		}		
		
	}	
	
	public boolean isActing() {
		return false;   // acting is handled by the woman (always only one)
	}	
			
}
